import sqlalchemy as sa
import worek

from worek.tests.helpers import PostgresDialectTestBase


class TestCorePGBackup(PostgresDialectTestBase):

    def test_backup_creates_full_restoreable_backup(self, tmpdir, pg_clean_engine):
        backup_file = tmpdir.join('test.backup.bin').strpath

        with open(backup_file, 'w+') as fp:
            worek.backup(fp, saengine=pg_clean_engine)

        with open(backup_file, 'rb') as fp:
            assert fp.read(5) == b'PGDMP'
            assert b'CREATE SCHEMA public' in fp.read()

        self.create_table(pg_clean_engine, 'should_be_removed')

        with open(backup_file, 'r') as fp:
            worek.restore(fp, saengine=pg_clean_engine)

        try:
            pg_clean_engine.execute('SELECT * FROM should_be_removed;')
        except sa.exc.ProgrammingError as e:
            assert 'relation "should_be_removed" does not exist' in str(e)



import worek.dialects.postgres as pgdialect
from worek.exc import WorekException


class WorekOperationException(WorekException):
    pass


def backup(backup_file, backup_type='full', **params):
    """Create backup of the database to the backup file

    :param backup_file: The file to send the backup to, this can be any file-like object including a
        a stream like. sys.stdin and sys.stdout should work no problem.
    :param backup_type: The type of database backup requested. The only option is 'full'.

    :param driver: the driver to use for connecting to the database
    :param host: the host of the database server
    :param port: the port of the database server
    :param user: the user of the database server
    :param password: the password of the database server
    :param dbname: the database name to backup
    :param saengine: an optional sqlalchemy engine, if this is passed, this will be used for the
        backup and other connection type parameters (e.g. driver, host, port) will be ignored
    :param version: version of PG client executables to use
    """
    PG = pgdialect.Postgres(
        engine=params.get('saengine') or pgdialect.Postgres.construct_engine_from_params(**params),
        schemas=params.get('schemas'),
        version=params.get('version'),
    )

    if not PG.engine_can_connect:
        raise WorekOperationException('Can\'t connect to the database.')

    if backup_type == 'full':
        PG.backup_binary(backup_file)
    else:
        raise NotImplementedError('Only full backups are available at this time.')


def restore(restore_file, file_format=None, clean_existing_database=True, **params):
    """Restore a backup file to the specified database

    :param restore_file: The file to pull the backup from, this can be any file-like object
        including a a stream like. sys.stdin and sys.stdout should work no problem.

    :param file_format: an optional file format. By default we try to be smart about this and detect
        the type of file, but sometimes we can't and this allows hard setting it.
    :param clean_existing_database: clean an existing database before restore
    :param driver: the driver to use for connecting to the database
    :param host: the host of the database server
    :param port: the port of the database server
    :param user: the user of the database server
    :param password: the password of the database server
    :param dbname: the database name to backup
    :param saengine: an optional sqlalchemy engine, if this is passed, this will be used for the
        backup and other connection type parameters (e.g. driver, host, port) will be ignored
    :param version: version of PG client executables to use
    """
    PG = pgdialect.Postgres(
        engine=params.get('saengine') or pgdialect.Postgres.construct_engine_from_params(**params),
        schemas=params.get('schemas'),
        version=params.get('version'),
    )

    if not PG.engine_can_connect:
        raise WorekOperationException('Can\'t connect to the database.')

    if clean_existing_database:
        PG.clean_existing_database()

    # perform the restore
    if file_format == 'c':
        return PG.restore_binary(restore_file)
    elif file_format == 't':
        return PG.restore_text(restore_file)
    elif file_format is None:
        return PG.restore(restore_file)
    else:
        raise NotImplementedError(
            'Got an unexpected file_format. {} is not a valid type, expecting'
            ' "c", "t", or nothing.'.format(file_format)
        )



import click
import worek.core as core


@click.group()
def cli():
    pass


@cli.command(help="Create a database backup")
@click.option('-h', '--host', default=None, help="connection hostname for server")
@click.option('-p', '--port', default=None, help="connection port for server")
@click.option('-u', '--user', default=None, help="connection username for server")
@click.option('-d', '--dbname', default=None, help="database to backup")
@click.option('-e', '--engine', default='postgres', help="database type [postgres]")
@click.option('-s', '--schema', multiple=True,
              help="schemas to backup up, can be used multiple times")
@click.option('-f', '--file', 'output_file', default=None, type=click.File(mode='w'),
              help="path to file backup location, otherwise pipe to STDOUT")
@click.option('-v', '--version', default=None, help="major version of PG client utilities")
def backup(host, port, user, dbname, engine, schema, output_file, version):
    file_name = output_file if output_file is not None else click.get_text_stream('stdout')

    try:
        core.backup(
            file_name,
            schemas=schema,
            host=host,
            port=port,
            user=user,
            dbname=dbname,
            version=version
        )
    except core.WorekOperationException as e:
        click.echo(str(e), err=True)


@cli.command(help="Restore a database backup")
@click.option('-h', '--host', default=None, help="connection hostname for server")
@click.option('-p', '--port', default=None, help="connection port for server")
@click.option('-u', '--user', default=None, help="connection username for server")
@click.option('-d', '--dbname', default=None, help="database to backup")
@click.option('-e', '--engine', default='postgres', help="database type [postgres]")
@click.option('-s', '--schema', multiple=True,
              help="schemas to backup up, can be used multiple times")
@click.option('-f', '--file', 'restore_file', default=None, type=click.File(mode='rb'),
              help="path to file backup location, otherwise the read from STDIN")
@click.option('-F', '--format', 'file_format', type=click.Choice(['c', 't']), default=None,
              help="backup file format, ([c]ustom, [t]ext)")
@click.option('-v', '--version', default=None, help="major version of PG client utilities")
def restore(host, port, user, dbname, engine, schema, restore_file, file_format, version):
    file_name = restore_file if restore_file is not None else click.get_binary_stream('stdin')

    if not restore_file and not file_format:
        raise click.BadArgumentUsage(
            'You must specify the file format (-F) when using a pipe to STDIN. We can not'
            ' automatically determine the file format when using a pipe.',
        )

    core.restore(file_name, schema=schema, host=host, port=port, user=user, dbname=dbname,
                 file_format=file_format, version=version)




class MockCLIExecutor:
    def __init__(self, returncode=0, stderr=b'', stdout=b''):
        self.returncode = returncode
        self.stderr = stderr
        self.stdout = stdout

    def __call__(self, *args, **kwargs):
        self.captured_args = args
        self.captured_kwargs = kwargs
        return self


class PostgresDialectTestBase():
    def create_extension(self, conn, ext, schema=None):
        schema = 'WITH SCHEMA "{}"'.format(schema) if schema else ''
        sql = 'CREATE EXTENSION IF NOT EXISTS "{}" {}'.format(ext, schema)
        conn.execute(sql)

    def create_table(self, conn, table, schema=None):
        with_schema = '.'.join([schema, table]) if schema else table
        sql = 'CREATE TABLE {} ( id integer PRIMARY KEY );'.format(with_schema)
        conn.execute(sql)

    def create_sequence(self, conn, sequence, schema=None):
        with_schema = '.'.join([schema, sequence]) if schema else sequence
        sql = 'CREATE SEQUENCE {};'.format(with_schema)
        conn.execute(sql)

    def create_function(self, conn, function, schema=None):
        with_schema = '.'.join([schema, function]) if schema else function
        sql = (
            '''
            CREATE OR REPLACE FUNCTION {}(int) RETURNS int
            AS $$ SELECT 1 $$ LANGUAGE SQL;
            '''.format(with_schema)
        )
        conn.execute(sql)

    def create_type(self, conn, type_, schema=None):
        with_schema = '.'.join([schema, type_]) if schema else type_
        sql = 'CREATE TYPE {};'.format(with_schema)
        conn.execute(sql)



import os


def latest_version(executable):
    try:
        # This is not technically correct, this just grabs the first it finds on the path
        return which(executable)[0]
    except IndexError:
        return None


def which(name, flags=os.X_OK):
    """
    Search PATH for executable files with the given name.


    .. note:: This function was copied from twisted. Find the original source
    here: https://github.com/twisted/twisted/blob/6ac66416/src/twisted/python/procutils.py and the
    associated license here: https://github.com/twisted/twisted/blob/6ac664/LICENSE (MIT)

    Copyright (c) 2001-2018 Twisted Matrix Laboratories


    On newer versions of MS-Windows, the PATHEXT environment variable will be set to the list of
    file extensions for files considered executable. This will normally include things like ".EXE".
    This function will also find files with the given name ending with any of these extensions.

    On MS-Windows the only flag that has any meaning is os.F_OK. Any other flags will be ignored.

    @type name: C{str}
    @param name: The name for which to search.

    @type flags: C{int}
    @param flags: Arguments to L{os.access}.

    @rtype: C{list}
    @param: A list of the full paths to files found, in the order in which they were found.
    """
    result = []
    exts = list(filter(None, os.environ.get('PATHEXT', '').split(os.pathsep)))
    path = os.environ.get('PATH', None)

    if path is None:
        return []

    for p in os.environ.get('PATH', '').split(os.pathsep):
        p = os.path.join(p, name)
        if os.access(p, flags):
            result.append(p)
        for e in exts:
            pext = p + e
            if os.access(pext, flags):
                result.append(pext)

    return result



import getpass
import os
from unittest import mock

import pytest
import sqlalchemy as sa


from worek.dialects.postgres import (
    Postgres as PG,
    PostgresCommand,
)
from worek.tests.helpers import (
    MockCLIExecutor,
    PostgresDialectTestBase,
)


class TestPostgresDialectInternals(PostgresDialectTestBase):
    def test_construct_engine_from_params_with_passed_values(self):
        engine = PG.construct_engine_from_params(
            user='testuser',
            password='testpassword',
            host='testhost',
            port='1111',
            dbname='testname'
        )
        assert engine.url.drivername == 'postgresql'
        assert engine.url.username == 'testuser'
        assert engine.url.password == 'testpassword'
        assert engine.url.host == 'testhost'
        assert engine.url.port == 1111
        assert engine.url.database == 'testname'

    def test_construct_engine_from_env(self, monkeypatch):
        monkeypatch.setenv('PGUSER', 'envuser')
        monkeypatch.setenv('PGPASSWORD', 'envpass')
        monkeypatch.setenv('PGHOST', 'envhost')
        monkeypatch.setenv('PGPORT', '1234')
        monkeypatch.setenv('PGDATABASE', 'envdb')
        engine = PG.construct_engine_from_params()

        assert engine.url.drivername == 'postgresql'
        assert engine.url.username == 'envuser'
        assert engine.url.password == 'envpass'
        assert engine.url.host == 'envhost'
        assert engine.url.port == 1234
        assert engine.url.database == 'envdb'

    def test_construct_engine_from_default(self, monkeypatch):
        with monkeypatch.context() as m:
            m.delenv('PGUSER', raising=False)
            m.delenv('PGPASSWORD', raising=False)
            m.delenv('PGHOST', raising=False)
            m.delenv('PGPORT', raising=False)
            m.delenv('PGDATABASE', raising=False)

            engine = PG.construct_engine_from_params()

        assert engine.url.drivername == 'postgresql'
        assert engine.url.username == getpass.getuser()
        assert engine.url.password == ''
        assert engine.url.host == 'localhost'
        assert engine.url.port == 5432
        assert engine.url.database == engine.url.username

    def test_autodetects_schemas(self, pg_unclean_engine, pg_uniqueschema):
        pg = PG(pg_unclean_engine)

        assert pg.schemas == ['public', pg_uniqueschema]

    def test_uses_passed_in_schemas(self):
        pg = PG(None, schemas=['other'])

        assert pg.schemas == ['other']

    def test_get_table_list_from_db(self, pg_unclean_engine, pg_uniqueschema):
        pg = PG(pg_unclean_engine)

        conn = pg_unclean_engine.connect()
        self.create_table(conn, 'testtbl_other')
        self.create_table(conn, 'testtbl', schema=pg_uniqueschema)

        assert pg.get_table_list_from_db(pg_uniqueschema) == ['testtbl']

    def test_get_seq_list_from_db(self, pg_unclean_engine, pg_uniqueschema):
        pg = PG(pg_unclean_engine)

        conn = pg_unclean_engine.connect()
        self.create_sequence(conn, 'testseq_other')
        self.create_sequence(conn, 'testseq', schema=pg_uniqueschema)

        assert pg.get_seq_list_from_db(pg_uniqueschema) == ['testseq']

    def test_get_function_list_from_db(self, pg_unclean_engine, pg_uniqueschema):
        pg = PG(pg_unclean_engine)

        conn = pg_unclean_engine.connect()
        self.create_function(conn, 'testfunc_other')
        self.create_function(conn, 'testfunc', schema=pg_uniqueschema)

        assert pg.get_function_list_from_db(pg_uniqueschema) == [('testfunc', 'integer')]

    def test_get_function_list_from_db_does_not_include_extensions(
            self, pg_unclean_engine, pg_uniqueschema):

        pg = PG(pg_unclean_engine)

        conn = pg_unclean_engine.connect()
        self.create_function(conn, 'testfunc_other')
        self.create_extension(conn, 'uuid-ossp', schema=pg_uniqueschema)
        self.create_function(conn, 'testfunc', schema=pg_uniqueschema)

        assert pg.get_function_list_from_db(pg_uniqueschema) == [('testfunc', 'integer')]

    def test_get_type_list_from_db(self, pg_unclean_engine, pg_uniqueschema):
        pg = PG(pg_unclean_engine)

        conn = pg_unclean_engine.connect()
        self.create_type(conn, 'testtype_other')
        self.create_type(conn, 'testtype', schema=pg_uniqueschema)

        assert pg.get_type_list_from_db(pg_uniqueschema) == ['testtype']

    def test_get_non_system_schemas_list_from_db(self, pg_unclean_engine, pg_uniqueschema):
        pg = PG(pg_unclean_engine)

        assert pg.get_non_system_schemas() == ['public', pg_uniqueschema]

    def test_drop_an_entire_schema(self, pg_unclean_engine, pg_uniqueschema):
        pg = PG(pg_unclean_engine)

        conn = pg_unclean_engine.connect()
        self.create_function(conn, 'testfunc', schema=pg_uniqueschema)
        self.create_sequence(conn, 'testseq', schema=pg_uniqueschema)
        self.create_table(conn, 'testtbl', schema=pg_uniqueschema)
        self.create_type(conn, 'testtype', schema=pg_uniqueschema)

        pg.drop_schema(pg_uniqueschema)

        assert pg.errors == []
        assert pg.get_function_list_from_db(pg_uniqueschema) == []
        assert pg.get_seq_list_from_db(pg_uniqueschema) == []
        assert pg.get_table_list_from_db(pg_uniqueschema) == []
        assert pg.get_type_list_from_db(pg_uniqueschema) == []

    def test_url_translation_to_cli_commands_with_schemas(self):
        executor = MockCLIExecutor()
        engine = sa.create_engine('postgresql://user:password@host:1111/dbname')
        pg = PG(engine, executor=executor, schemas=['public', 'other'], version='10')
        pg._execute_cli_command(PostgresCommand.BACKUP)

        assert executor.captured_args[0][0].endswith('pg_dump')
        assert set(executor.captured_args[0][1:]) == {
            '--dbname=dbname',
            '--host=host',
            '--schema=public',
            '--schema=other',
            '--username=user',
            '--port=1111',
        }
        assert executor.captured_kwargs['env']['PGPASSWORD'] == 'password'
        assert executor.captured_kwargs['env']['PGCLUSTER'] == '10/localhost:'

    def test_command_doesnt_use_schema_for_text_restore(self):
        executor = MockCLIExecutor()
        engine = sa.create_engine('postgresql://user:password@host:1111/dbname')
        pg = PG(engine, executor=executor, schemas=['public', 'other'], version='10')
        pg._execute_cli_command(PostgresCommand.RESTORE_TEXT)

        assert executor.captured_args[0][0].endswith('psql')
        assert set(executor.captured_args[0][1:]) == {
            '--dbname=dbname',
            '--host=host',
            '--username=user',
            '--port=1111',
        }
        assert executor.captured_kwargs['env']['PGPASSWORD'] == 'password'
        assert executor.captured_kwargs['env']['PGCLUSTER'] == '10/localhost:'

    @pytest.mark.skipif(not os.environ.get('PGVERSION'),
                        reason='Must set PGVERSION to server version')
    def test_command_uses_correct_exe_version(self, pg_unclean_engine):
        executor = MockCLIExecutor()
        expected_version = os.environ.get('PGVERSION')

        # Default
        pg = PG(pg_unclean_engine, executor=executor, schemas=['public', 'other'])
        pg._execute_cli_command(PostgresCommand.RESTORE_TEXT)
        assert executor.captured_kwargs['env']['PGCLUSTER'] == '{}/localhost:'.format(
            expected_version)

        # Explicit version
        pg = PG(pg_unclean_engine, executor=executor, schemas=['public', 'other'], version='11')
        pg._execute_cli_command(PostgresCommand.RESTORE_TEXT)
        assert executor.captured_kwargs['env']['PGCLUSTER'] == '11/localhost:'

        pg = PG(pg_unclean_engine, executor=executor, schemas=['public', 'other'], version='10')
        pg._execute_cli_command(PostgresCommand.RESTORE_TEXT)
        assert executor.captured_kwargs['env']['PGCLUSTER'] == '10/localhost:'

        pg = PG(pg_unclean_engine, executor=executor, schemas=['public', 'other'], version='9')
        pg._execute_cli_command(PostgresCommand.RESTORE_TEXT)
        assert executor.captured_kwargs['env']['PGCLUSTER'] == '9/localhost:'

    def test_command_pg_wrapper_not_installed(self):
        executor = MockCLIExecutor()
        engine = sa.create_engine('postgresql://user:password@host:1111/dbname')
        pg = PG(engine, executor=executor, schemas=['public', 'other'], version='10')

        with mock.patch('shutil.which') as m_which:
            m_which.return_value = None
            with pytest.raises(OSError, match='pg_wrapper is required if a version is specified'):
                pg._execute_cli_command(PostgresCommand.RESTORE_TEXT)
            m_which.assert_called_once_with('psql')

        with mock.patch('os.path.islink') as m_is_link:
            m_is_link.return_value = False
            with pytest.raises(OSError, match='pg_wrapper is required if a version is specified'):
                pg._execute_cli_command(PostgresCommand.RESTORE_TEXT)
            m_is_link.assert_called_once_with('/usr/bin/psql')

        with mock.patch('os.path.realpath') as m_real_path:
            m_real_path.return_value = '/foo/bar'
            with pytest.raises(OSError, match='pg_wrapper is required if a version is specified'):
                pg._execute_cli_command(PostgresCommand.RESTORE_TEXT)
            m_real_path.assert_called_once_with('/usr/bin/psql')


class TestPostgresDialectBackup(PostgresDialectTestBase):

    def test_backup_creates_basic_sql_backup(self, tmpdir, pg_unclean_engine, pg_uniqueschema):
        pg = PG(pg_unclean_engine, schemas=[pg_uniqueschema])
        backup_dir = tmpdir.join('/test.backup').strpath

        with open(backup_dir, mode='w+') as fp:
            pg.backup_text(fp)
            fp.seek(0)
            result = fp.read()

        assert 'CREATE SCHEMA {};'.format(pg_uniqueschema) in result
        assert 'CREATE SCHEMA public;' not in result


class TestPostgresDialectIntegration(PostgresDialectTestBase):

    def test_create_and_restore_database_text_backup(self, tmpdir, pg_clean_engine):
        pg = PG(pg_clean_engine)
        backup_file = tmpdir.join('test.backup.text').strpath

        self.create_table(pg_clean_engine, 'testtbl')
        pg_clean_engine.execute('INSERT INTO testtbl VALUES (1),(2),(3)')

        with open(backup_file, mode='w+') as fp:
            pg.backup_text(fp)

        pg_clean_engine.execute('DROP TABLE testtbl;')

        try:
            pg_clean_engine.execute('SELECT * FROM testtbl;')
        except sa.exc.ProgrammingError as e:
            if 'relation "testtbl" does not exist' not in str(e):
                raise

        with open(backup_file, mode='rb') as fp:
            command_result = pg.restore_text(fp)

        assert command_result.returncode == 0

        result = pg_clean_engine.execute('SELECT * FROM testtbl;')
        assert result.rowcount == 3
        assert set([x[0] for x in result.fetchall()]) == {1, 2, 3}

    def test_create_and_restore_database_binary_backup(self, tmpdir, pg_clean_engine):
        pg = PG(pg_clean_engine)
        backup_file = tmpdir.join('test.backup.bin').strpath

        self.create_table(pg_clean_engine, 'testtbl')
        pg_clean_engine.execute('INSERT INTO testtbl VALUES (1),(2),(3)')

        with open(backup_file, mode='w+') as fp:
            pg.backup_binary(fp)

        pg_clean_engine.execute('DROP TABLE testtbl;')

        try:
            pg_clean_engine.execute('SELECT * FROM testtbl;')
        except sa.exc.ProgrammingError as e:
            if 'relation "testtbl" does not exist' not in str(e):
                raise

        with open(backup_file, mode='rb') as fp:
            command_result = pg.restore(fp)

        assert command_result.returncode == 0

        result = pg_clean_engine.execute('SELECT * FROM testtbl;')
        assert result.rowcount == 3
        assert set([x[0] for x in result.fetchall()]) == {1, 2, 3}



from click.testing import CliRunner
import sqlalchemy as sa

from worek.tests.helpers import PostgresDialectTestBase
from worek.cli import cli


class TestCLIPGBackup(PostgresDialectTestBase):

    def engine_to_cli_params(self, engine):
        return [
            '--host', engine.url.host,
            '--user', engine.url.username,
            '--port', engine.url.port,
            '--dbname', engine.url.database,
        ]

    def test_cli_can_create_restoreable_backup(self, tmpdir, pg_clean_engine):
        backup_file = tmpdir.join('test.backup.bin').strpath

        runner = CliRunner()
        result = runner.invoke(
            cli, ['backup', '-f', backup_file] + self.engine_to_cli_params(pg_clean_engine)
        )

        assert result.exit_code == 0

        with open(backup_file, 'rb') as fp:
            assert fp.read(5) == b'PGDMP'
            data = fp.read()
            assert b'CREATE SCHEMA public' in data

        self.create_table(pg_clean_engine, 'manualtest')

        assert pg_clean_engine.execute('SELECT * from manualtest')

        result = runner.invoke(
            cli,
            ['restore', '-f', backup_file] + self.engine_to_cli_params(pg_clean_engine),
        )

        assert result.exit_code == 0

        try:
            assert pg_clean_engine.execute('SELECT * from manualtest')
        except sa.exc.ProgrammingError as e:
            assert 'relation "manualtest" does not exist' in str(e)



import enum
import getpass
import logging
import os
import re
import shutil
import subprocess

import sqlalchemy as sa
import psycopg2

from worek.exc import WorekException


log = logging.getLogger(__name__)


class WorekPostgresError(WorekException):
    pass


class PostgresCLIError(WorekPostgresError):
    def __init__(self, command_result):
        self.command_result = command_result


class PostgresInputError(WorekPostgresError):
    pass


class PostgresCommand(enum.Enum):
    BACKUP = 'pg_dump'
    RESTORE_BINARY = 'pg_restore'
    RESTORE_TEXT = 'psql'


class Postgres:
    """A logical backup / restore implementation for Postgres databases.

    Utilizes the pg_dump and pg_restore commands and some extra features to allow database users to
    manage backups and restores of their own data.
    """

    def __init__(self, engine, schemas=None, executor=None, version=None):
        """Postgres Database Management Tool

        :param engine: a SQLAlchemy Engine which connects to the relevant database
        :param schemas: schemas you want to restore or backup
        :param executor: the method for executing the PG commands against the database. By default
            this uses `subprocess.run` but you can use the `tests.helpers.MockCLIExecutor` or any
            callable that accepts a list of CLI arguments and `**kwargs`.
        :param version: the version of PG client executables to use

        .. note:: `schemas` is ignored when working with a RESTORE_TEXT operation because postgres
            can not selectively restore a plain text backup. Use a binary backup if you want to
            manage schemas manually.
        """
        self.engine = engine
        self._schemas = schemas

        self.errors = []
        self.executor = subprocess.run if executor is None else executor
        self.version = version

    @property
    def schemas(self):
        if not self._schemas:
            self._schemas = self.get_non_system_schemas()

        return self._schemas

    def _default_version(self):
        sql = sa.select([sa.func.current_setting('server_version')])
        result = self.engine.execute(sql).fetchone()

        match = re.match(r'(\d+\.\d+)', result[0])
        version = match.group(1)
        if version.startswith('9'):
            return version
        return version.split('.')[0]

    @property
    def engine_can_connect(self):
        try:
            self.engine.execute("SELECT 1")
            return True
        except psycopg2.Error:
            return False

    @classmethod
    def construct_engine_from_params(cls, **params):
        """Create a SQLAlchemy Engine from the passed params and sensible defaults


        .. note:: Because we have to clean the database manually when running as a non-superuser
            (can't drop the database) we have to have a fully resolved database URL for SQLAlchemy.
            The Postgres commands will use the same URL params in the execution of the cli commands
        """

        driver = params.get('driver') or 'postgresql'

        if 'postgresql' not in driver:
            raise ValueError('The database driver must be for postgresql, got {}.'.format(driver))

        user = params.get('user') or os.environ.get('PGUSER') or getpass.getuser()
        password = params.get('password') or os.environ.get('PGPASSWORD', '')
        host = params.get('host') or os.environ.get('PGHOST', 'localhost')
        port = params.get('port') or os.environ.get('PGPORT', 5432)
        dbname = params.get('dbname') or os.environ.get('PGDATABASE', user)

        return sa.create_engine(
            'postgresql://{user}:{password}@{host}:{port}/{dbname}'.format(
                user=user,
                password=password,
                host=host,
                port=port,
                dbname=dbname
            )
        )

    @classmethod
    def cli_flags_for_url(cls, url):
        flags = []

        if url.username:
            flags += ['--username={}'.format(url.username)]

        if url.host:
            flags += ['--host={}'.format(url.host)]

        if url.port:
            flags += ['--port={}'.format(url.port)]

        if url.database:
            flags += ['--dbname={}'.format(url.database)]

        return flags

    def _pg_wrapper_cluster(self, exe_name):
        if not is_pg_wrapper_available(exe_name) and self.version:
            raise OSError('pg_wrapper is required if a version is specified')
        return '{}/localhost:'.format(self.version or self._default_version())

    def _execute_cli_command(self, command, additional_args=None, stdin=None, stdout=None,
                             stderr=None):
        """
        Execute a CLI Command putting STDOUT into output.

        .. note:: While you could pass `--file` in `command_args` for pg_dump and pg_restore, it is
        much more flexible to allow those programs to dump to STDOUT and then use `subprocess` to
        pipe the data into the file you choose. This allows neat things like passing in sys.stdout
        so that it can get all the way to the console or passed to GZIP or an encryption protocol,
        all without having to save to disk first.

        :param command: the postgres command you want to run, should be an instance of
            `PostgresCommand`
        :param additional_args: a list of arguments which should be passed to the CLI command.
        :param stdin: a stream or pipe to use for standard input, useful when you want to pipe a
            restore command from some other tool like gzip or openssl.
        :param stdout: a stream or pipe to use for standard output, useful when you want to pipe a
            backup command to some other tool like gzip or openssl.
        :param stderr: a stream or pipe to use for standard error, by default this is just a
            subprocess pipe.
        """
        stdin = stdin or subprocess.PIPE
        stdout = stdout or subprocess.PIPE
        stderr = stderr or subprocess.PIPE

        if not isinstance(command, PostgresCommand):
            raise NotImplementedError(
                'Unknown postgresql command. Are you using the PostgresCommand enum.'
            )

        env = os.environ.copy()
        url = self.engine.url

        cli_args = [command.value] + self.cli_flags_for_url(url)

        if command != PostgresCommand.RESTORE_TEXT:
            cli_args += ['--schema={}'.format(x) for x in self.schemas]

        env['PGCLUSTER'] = self._pg_wrapper_cluster(command.value)
        if url.password:
            env['PGPASSWORD'] = url.password

        cli_args.extend(additional_args or [])

        result = self.executor(cli_args, env=env, stdin=stdin, stdout=stdout, stderr=stderr)

        if result.returncode != 0:
            raise PostgresCLIError(result)

        return result

    def drop_schema(self, schema):
        for funcname, funcargs in self.get_function_list_from_db(schema):
            try:
                sql = 'DROP FUNCTION "{}"."{}" ({}) CASCADE'.format(schema, funcname, funcargs)
                self.engine.execute(sql)
            except Exception:
                raise

        for table in self.get_table_list_from_db(schema):
            try:
                self.engine.execute('DROP TABLE "{}"."{}" CASCADE'.format(schema, table))
            except Exception:
                raise

        for seq in self.get_seq_list_from_db(schema):
            try:
                self.engine.execute('DROP SEQUENCE "{}"."{}" CASCADE'.format(schema, seq))
            except Exception:
                raise

        for dbtype in self.get_type_list_from_db(schema):
            try:
                self.engine.execute('DROP TYPE "{}"."{}" CASCADE'.format(schema, dbtype))
            except Exception:
                raise

    def get_function_list_from_db(self, schema):
        """Returns a list of functions not associated with an extension

        .. note:: original versions of this function would return all functions even if they were
            related to installed extensions. Since our app users can't install extensions, we need
            to leave those function intact along with the extensions that installed them.
        """
        sql = """
            SELECT
                PR.proname,
                pg_get_function_identity_arguments(PR.oid) AS params
            FROM
                pg_proc PR
                JOIN pg_namespace NS ON NS.oid = PR.pronamespace
                LEFT JOIN pg_depend D ON D.objid = PR.oid
                AND D.deptype = 'e'
            WHERE
                NS.nspname = '{schema}'
                AND D.objid IS NULL;
        """.format(schema=schema)

        return [row for row in self.engine.execute(sql)]

    def get_table_list_from_db(self, schema):
        """
        Return a list of table names from the passed schema
        """

        sql = """
            SELECT table_name
            FROM information_schema.tables
            WHERE table_schema='{schema}';
        """.format(schema=schema)

        return [name for (name, ) in self.engine.execute(sql)]

    def get_seq_list_from_db(self, schema):
        """return a list of the sequence names from the current
           databases public schema
        """
        sql = """
            SELECT sequence_name
            FROM information_schema.sequences
            WHERE sequence_schema='{schema}';
        """.format(schema=schema)

        return [name for (name, ) in self.engine.execute(sql)]

    def get_non_system_schemas(self):
        """
        Return a list of non-postgres default schema

        .. note:: the use of pg_%, find the relevant documenation on determining system schemas
            here: https://www.postgresql.org/docs/10/static/ddl-schemas.html#DDL-SCHEMAS-CATALOG
        """

        all_non_default_schemas_query = """
            SELECT schema_name
            FROM information_schema.schemata
            WHERE
                "schema_name" NOT LIKE 'pg_%%'
            AND "schema_name" != 'information_schema';
        """

        result = self.engine.execute(all_non_default_schemas_query)

        return [x[0] for x in result.fetchall()]

    def get_type_list_from_db(self, schema):
        """return a list of the sequence names from the passed schema
        """
        sql = """
            SELECT t.typname as type
            FROM pg_type t
            LEFT JOIN pg_catalog.pg_namespace n
                ON n.oid = t.typnamespace
            WHERE
                ( t.typrelid = 0 OR
                    (
                        SELECT c.relkind = 'c'
                        FROM pg_catalog.pg_class c
                        WHERE c.oid = t.typrelid
                    )
                )
                AND NOT EXISTS (
                    SELECT 1
                    FROM pg_catalog.pg_type el
                    WHERE el.oid = t.typelem
                        AND el.typarray = t.oid
                )
                AND n.nspname = '{}'
        """.format(schema)

        return [name for (name, ) in self.engine.execute(sql)]

    def clean_existing_database(self, schemas=None):
        schemas = schemas if schemas is not None else self.schemas

        for schema in schemas:
            self.drop_schema(schema)

    def restore(self, buf, **kwargs):
        """Perform a "smart" restore, properly choosing between a binary or text version"""
        return self.restore_binary(buf, **kwargs)

        try:
            header = buf.peek(5)
        except AttributeError:
            raise PostgresInputError(
                'You must use a peekable stream when trying to automatically detect the backup file'
                ' type. If you are pipe data into worek from the CLI, you should explicitly set the'
                ' backup file type that you are piping to the restore command.')

        if header[:5] in ('PGDMP', b'PGDMP'):
            return self.restore_binary(buf, **kwargs)
        else:
            return self.restore_text(buf, **kwargs)

    def restore_binary(self, buf, no_owner=True, no_privileges=True, **kwargs):
        """Restore a binary backup from the passed buf

        :param buf: the buffer with the backup to restore
        :param no_owner: do not keep the owner information from the backup (default: True)
        :param no_privileges: do not restore privileges information from the backup (default: True)

        .. note:: Depending on the `self.executor`, the options for `buf` depend on the supported
            values. By default this class uses `subprocess.run` to execute the restore command, so
            you can pass anything to `buf` that the `stdin` argument would take for that function
            (i.e. PIPE, file, DEVNULL)
        """
        command_args = (
            []
            + (['--no-owner' if no_owner else ''])
            + (['--no-privileges' if no_privileges else ''])
        )
        return self._execute_cli_command(PostgresCommand.RESTORE_BINARY, command_args, stdin=buf)

    def restore_text(self, buf, **kwargs):
        """Restore a plain text backup from the passed buf

        :param buf: the buffer with the backup to restore

        .. note:: Depending on the `self.executor`, the options for `buf` depend on the supported
            values. By default this class uses `subprocess.run` to execute the restore command, so
            you can pass anything to `buf` that the `stdin` argument would take for that function
            (i.e. PIPE, file, DEVNULL)
        """
        return self._execute_cli_command(PostgresCommand.RESTORE_TEXT, [], stdin=buf)

    def backup_binary(self, buf, blobs=True):
        """Create a binary (--format=custom) backup of the postgres context

        :param buf: the buffer to store the results of the backup
        :param blobs: include blob data in backup (default: True)

        .. note:: Depending on the `self.executor`, the options for `buf` depend on the supported
            values. By default this class uses `subprocess.run` to execute the backup command, so
            you can pass anything to `buf` that the `stdout` argument would take for that function
            (i.e. PIPE, file, DEVNULL)
        """
        command_args = ['--format', 'custom'] + (['--blobs'] if blobs else [])
        return self._execute_cli_command(PostgresCommand.BACKUP, command_args, stdout=buf)

    def backup_text(self, buf):
        """Create a plain text backup of the postgres context

        :param buf: the buffer to store the results of the backup

        .. note:: Depending on the `self.executor`, the options for `buf` depend on the support
            values. By default this class uses `subprocess.run` to execute the backup command, so
            you can pass anything to `buf` that the `stdout` argument would take for that function
            (i.e. PIPE, file, DEVNULL)
        """

        command_args = ['--format', 'plain']
        return self._execute_cli_command(PostgresCommand.BACKUP, command_args, stdout=buf)


def is_pg_wrapper_available(exe_name):
    path = shutil.which(exe_name)
    if path is None:
        return False

    if not os.path.islink(path):
        return False
    real_path = os.path.realpath(path)
    exe_name = os.path.basename(real_path)
    return exe_name == 'pg_wrapper'



import random
import string

import sqlalchemy as sa
import pytest
from worek.dialects.postgres import Postgres as PG

DBNAME = 'worek-tests'


def pytest_configure(config):
    # clean / create the database for this run
    clean_database()


@pytest.fixture(scope='function')
def pg_unclean_engine():
    """Return a handle to the test database

    Use this fixture when you can reuse the DB across tests and it doesn't matter what else is in
    there. Try to use this fixture as often as possible since it is expensive to drop and create a
    clean database.
    """
    engine = PG.construct_engine_from_params(dbname=DBNAME)

    try:
        testconn = engine.connect()
        testconn.close()
    except sa.exc.OperationalError:
        clean_database()

    return engine


@pytest.fixture(scope='function')
def pg_clean_engine():
    clean_database()
    return PG.construct_engine_from_params(dbname=DBNAME)


def clean_database():
    """Return a handle to the test database

    :param admin_engine: you need to have a handle to the database with superuser abilities so you
        can drop and create the database

    Use this fixture when you need the database absolutely fresh, no data from other tests, be them
    schemas, functions, or otherwise, will be present. Note, this drops and creates the database, so
    try to only use this when absolutely necessary.
    """

    admin_engine = PG.construct_engine_from_params(dbname='postgres')
    local_conn = admin_engine.connect()

    # This bumps us out of a transaction because a DROP/CREATE command can't run in a transaction
    local_conn.execute('COMMIT')

    local_conn.execute('''
        SELECT pg_terminate_backend(pg_stat_activity.pid)
        FROM pg_stat_activity
        WHERE datname = '{}'
        AND pid <> pg_backend_pid();
    '''.format(DBNAME))
    local_conn.execute('DROP DATABASE IF EXISTS "{}";'.format(DBNAME))

    local_conn.execute('COMMIT')
    local_conn.execute('CREATE DATABASE "{}";'.format(DBNAME))


@pytest.fixture(scope='function')
def pg_uniqueschema(pg_unclean_engine):
    """Create and return the name of a unique schema, which will be cleaned up after the test

    """
    schema_name = ''.join(random.sample(string.ascii_lowercase, k=20))

    conn = pg_unclean_engine.connect()
    conn.execute('CREATE SCHEMA {}'.format(schema_name))
    conn.close()

    yield schema_name

    conn = pg_unclean_engine.connect()
    conn.execute('DROP SCHEMA IF EXISTS {} CASCADE'.format(schema_name))
    conn.close()
